def zeros_matrix(rows, cols):
    """
    Creates a matrix filled with zeros.
        :param rows: the number of rows the matrix should have
        :param cols: the number of columns the matrix should have
 
        :return: list of lists that form the matrix
    """
    M = []
    while len(M) < rows:
        M.append([])
        while len(M[-1]) < cols:
            M[-1].append(0.0)
 
    return M

def matrix_addition(A, B):
    """
    Adds two matrices and returns the sum
        :param A: The first matrix
        :param B: The second matrix
 
        :return: Matrix sum
    """
    # Section 1: Ensure dimensions are valid for matrix addition
    rowsA = len(A)
    colsA = len(A[0])
    rowsB = len(B)
    colsB = len(B[0])
    if rowsA != rowsB or colsA != colsB:
        raise ArithmeticError('Matrices are NOT the same size.')
 
    # Section 2: Create a new matrix for the matrix sum
    C = zeros_matrix(rowsA, colsB)
 
    # Section 3: Perform element by element sum
    for i in range(rowsA):
        for j in range(colsB):
            C[i][j] = A[i][j] + B[i][j]
 
    return C

def matrix_subtraction(A, B):
    """
    Subtracts matrix B from matrix A and returns difference
        :param A: The first matrix
        :param B: The second matrix
 
        :return: Matrix difference
    """
    # Section 1: Ensure dimensions are valid for matrix subtraction
    rowsA = len(A)
    colsA = len(A[0])
    rowsB = len(B)
    colsB = len(B[0])
    if rowsA != rowsB or colsA != colsB:
        raise ArithmeticError('Matrices are NOT the same size.')
 
    # Section 2: Create a new matrix for the matrix difference
    C = zeros_matrix(rowsA, colsB)
 
    # Section 3: Perform element by element subtraction
    for i in range(rowsA):
        for j in range(colsB):
            C[i][j] = A[i][j] - B[i][j]
 
    return C

def rate_of_change(series):
    res = matrix_subtraction([series[1:]], [series[0:-1]])
    return sum(res[0])/len(res[0])


from rpi2mqtt.switch import BasicSwitch, Switch
from rpi2mqtt.base import Sensor
from rpi2mqtt.mqtt import MQTT
from rpi2mqtt.temperature import BME280
from rpi2mqtt.config import Config
import RPi.GPIO as GPIO
import pendulum
import logging
import json
import rpi2mqtt.math as math
import time
from collections import deque


class HVAC(object):
    HEAT_PUMP = {
        'fan': 18,
        'compressor': 23,
        'reversing_valve': 12,
        'aux': 16,
    }

    HEAT_PUMP_MODES = {
        'off': [],
        'fan': [HEAT_PUMP['fan']],
        'heat': [HEAT_PUMP['fan'], HEAT_PUMP['compressor']],
        '_cool': [HEAT_PUMP['reversing_valve']],
        'cool': [HEAT_PUMP['fan'], HEAT_PUMP['compressor'], HEAT_PUMP['reversing_valve']],
        'aux': [HEAT_PUMP['fan'], HEAT_PUMP['compressor'], HEAT_PUMP['aux']],
        'boost': [HEAT_PUMP['aux']],
        'emergency': [HEAT_PUMP['fan'], HEAT_PUMP['aux']],
    }

    # ON = 'ON'
    # OFF = 'OFF'

    HEAT = 'heat'
    COOL = 'cool'
    AUX = 'aux'
    EMERGENCY = 'emergency'
    BOOST = 'boost'
    FAN = 'fan'
    FAN_ON = 'high'
    AUTO = 'auto'
    ON = 'on'
    OFF = 'off'


class HvacException(Exception):
    pass


class HestiaPi(Sensor):

    def __init__(self, **kwargs):
        # self._modes = HVAC.HEAT_PUMP_MODES
        super(HestiaPi, self).__init__(kwargs.get('name'), None, kwargs.get('topic'), 'climate', 'HestiaPi')
        self.mode = kwargs.get('mode', 'heat')
        # self.active = False
        # self.desired_mode = 'off'
        self.active_start_time = None
        self.set_point_cool = kwargs.get('cool_setpoint', 76)
        self.set_point_heat = kwargs.get('heat_setpoint', 68)
        # how much wiggle room in temperature reading before starting/stopping HVAC.
        # setting this too low can trigger frequence HVAC cycles.
        self.set_point_tolerance = kwargs.get('set_point_tolerance', 0.5)
        # Minimum time HVAC should run (in minutes)
        self.min_run_time = kwargs.get('min_run_time', 15)
        # how soon can HVAC be activated again after stopping (in minutes)
        self.min_trigger_cooldown_time = kwargs.get('min_trigger_cooldown_time', 15)
        self.last_mode_change_time = None
        self.last_hvac_state_change_time = None
        self.bme280 = None
        # container to holder mode switches. Do not use directly.
        self._modes = {}
        # container to store temperature history
        self.temperature_history = deque(maxlen=kwargs.get('temperature_history_period', 6))
        # Minimum temperature rate of change over 4 measurements
        self.minimum_temp_rate_of_change = kwargs.get('minimum_temp_rate_of_change', -0.25)
        # super(HestiaPi, self).__init__(name, None, topic, 'climate', 'HestiaPi')
        # put thermostat into test mode. i.e. don't trigger HVAC commands
        self.dry_run = kwargs.get('dry_run')
        # save boost state
        self._boosting_heat = HVAC.OFF
        self._boosting_start_time = None
        self._boosting_enabled_switch = None
        self.aux_enabled = kwargs.get('aux_enabled', HVAC.ON)

        self.setup()

    def setup(self):
        logging.debug('Setting up HestiaPi')
        self.bme280 = BME280(self.name, self.topic)
        self._boosting_enabled_switch = HVACAuxSwitch(self.name, None, self.topic, self.aux_enabled)

        for mode, pins in HVAC.HEAT_PUMP_MODES.items():
            switch = BasicSwitch(self.name, pins, '{}_{}'.format(self.topic, mode), mode)
            # switch.setup()
            self._modes[mode] = switch

        # setup GPIO inputs on HVAC pins
        GPIO.setmode(GPIO.BCM)
        for capability, pin in HVAC.HEAT_PUMP.items():
            GPIO.setup(pin, GPIO.IN)

        # Subscribe to MQTT command topics
        MQTT.subscribe(self.mode_command_topic, self.mqtt_set_mode_callback)
        MQTT.subscribe(self.temperature_set_point_command_topic, self.mqtt_set_temperature_set_point_callback)
        MQTT.subscribe(self.fan_command_topic, self.mqtt_set_fan_state_callback)
        MQTT.subscribe(self.aux_command_topic, self.mqtt_set_aux_mode_callback)

    @property
    def mode_command_topic(self):
        return '{}/mode/set'.format(self.topic)

    @property
    def temperature_set_point_command_topic(self):
        return '{}/temperature/set'.format(self.topic)

    @property
    def fan_command_topic(self):
        return '{}/fan/set'.format(self.topic)

    @property
    def  aux_command_topic(self):
        return '{}/aux/set'.format(self.topic)

    @property
    def homeassistant_mqtt_config_topic(self):
        return 'homeassistant/{}/{}/config'.format('climate', self.name)

    @property
    def homeassistant_mqtt_config(self):
        return {
                'name': '{}_{}'.format(self.name, self.device_class),
                'unique_id': '{}_{}_{}_rpi2mqtt'.format(self.name, self.device_model, self.device_class),
                "json_attributes_topic": self.topic,
                'device': self.device_config,
                'min_temp': 65,
                'max_temp': 80,
                'initial': 72,
                'modes': ['off', 'auto', 'heat', 'cool', 'aux'],
                'fan_modes': ['auto','high'],
                'action_topic': self.topic,
                'action_template': '{{ value_json.hvac_state }}',
                'current_temperature_topic': self.topic,
                'current_temperature_template': '{{ value_json.current_temperature | round(1) }}',
                'mode_state_topic': self.topic,
                'mode_state_template': '{{ value_json.mode }}',
                'mode_command_topic': self.mode_command_topic, 
                'temperature_state_topic': self.topic,
                'temperature_state_template': '{{ value_json.set_point }}',
                'temperature_command_topic': self.temperature_set_point_command_topic,
                'aux_state_topic': self.topic,
                'aux_state_topic': '{{ value_json.aux_mode }}', # TODO refactor aux_mode to aux_state
                'aux_command_topic': self.aux_command_topic,
                'fan_modes': ['auto', 'high'],
                'fan_mode_state_topic': self.topic,
                'fan_mode_state_template': '{{ value_json.fan_state }}',
                'fan_mode_command_topic': self.fan_command_topic
            }

    def set_state(self, mode, state):
        if not self.dry_run:
            if state == HVAC.ON:
                if mode not in [HVAC.FAN, HVAC.BOOST]:
                    self.active_start_time = pendulum.now()

                if mode == HVAC.COOL:
                    self._modes['heat'].on()
                    # delay reversing valve turn on before cooling starts
                    time.sleep(15)
                    self._modes['_cool'].on()
                else:
                    self._modes[mode].on()
                    
                # confirm mode change
                if mode == self.hvac_state: # TODO if boosting then only check boosting pin is active
                    logging.debug('Turned {} {}.'.format(mode, state))
                else:
                    logging.warn('Did not set HVAC state to {}. Try again.'.format(mode))

            elif state == HVAC.OFF:
                self._modes[mode].off()
                
                # if mode == HVAC.COOL:
                #     time.sleep(3)
                #     self._modes['_cool'].off()

                if mode not in [HVAC.FAN, HVAC.BOOST]:
                    self.active_start_time = None
                    self.temperature_history = []

                # confirm mode change
                if 'off' == self.hvac_state:
                    logging.debug('Turned {} {}.'.format(mode, state))
                else:
                    logging.warn('Did not set HVAC state to {}. Try again.'.format(mode))
            else:
                raise HvacException("State '{}' is not a valid state.".format(state))
        if mode not in [HVAC.FAN, HVAC.BOOST]:
            self.last_hvac_state_change_time = pendulum.now()

    @property
    def active_time(self):
        if self.active:
            try:
                return (pendulum.now() - self.active_start_time).in_minutes() 
            except Exception as e:
                logging.exception(e)
        return 0

    @property
    def boosting_active_time(self):
        if self._boosting_heat == HVAC.ON:
            try:
                return (pendulum.now() - self._boosting_start_time).in_minutes() 
            except Exception as e:
                logging.exception(e)
        return 0

    @property
    def active(self):
        return self.hvac_state in  [HVAC.COOL, HVAC.HEAT, HVAC.AUX]

    @property
    def minutes_since_last_mode_change(self):
        # if self.active:
        if self.last_mode_change_time:
            _minutes = (pendulum.now() - self.last_mode_change_time).in_minutes() 
            logging.info(f'{_minutes} minutes since last mode change.')
            return _minutes
        else:
            return 1000

    @property
    def minutes_since_last_hvac_state_change(self):
        if self.last_hvac_state_change_time:
            return (pendulum.now() - self.last_hvac_state_change_time).in_minutes() 
        else:
            return 1000

    @property
    def set_point(self):
        if self.mode == HVAC.HEAT:
            return self.set_point_heat
        elif self.mode == HVAC.COOL:
            return self.set_point_cool

    @property
    def hvac_state(self):
        """Current HVAC mode based on active GPIO pins."""
        active_pins = []
        # read pin state
        for capability, pin in HVAC.HEAT_PUMP.items():
            if GPIO.input(pin):
                active_pins.append(pin)
        active_pins = set(active_pins)

        # search heat pump modes for a match
        for mode, p in HVAC.HEAT_PUMP_MODES.items():
            if active_pins == set(p):
                logging.debug('HVAC state is "{}". Active GPIO pins = {}'.format(mode, active_pins))
                return mode

    @property
    def fan_state(self):
        if self.hvac_state == HVAC.FAN:
            return HVAC.FAN_ON
        else:
            return HVAC.AUTO

    @property
    def ha_hvac_state(self):
        state = self.hvac_state
        if state in ['heat','cool']:
            return '{}ing'.format(state)
        elif state == HVAC.AUX:
            return 'heating'
        else:
            return state

    @property
    def current_temperature(self):
        return round(self.temperature)

    @property
    def temperature(self):
        return self.bme280.state()['temperature']

    def append_tempearture_history(self):
        """Save current temperature in _temperature history and maintain N readings."""
         # if system is active log temperature changes for analysis
        # if self.active:
        self.temperature_history.append(self.temperature)
        logging.debug('Temperature history = {}'.format(self.temperature_history))
        # if len(self.temperature_history) > 6: # how many readings should we keep track of. 4 is ~20 minutes.
        #     self.temperature_history.pop(0)

    @property
    def temperature_rate_of_change(self):
        if len(self.temperature_history) > 1:
            roc = math.rate_of_change(self.temperature_history)
            logging.debug('Temperature rate of change is {}.'.format(roc))
            return roc

    def state(self):
        data = self.bme280.state()
        # aux_enabled_state = self._boosting_enabled_switch.state()
        return {
            'bme280': data,
            'mode': self.mode,
            'aux_enabled': self.aux_enabled,
            'aux_mode': self._boosting_heat,
            'active_time': self.active_time,
            'aux_active_time': self.boosting_active_time,
            'active': self.active,
            'hvac_state': self.ha_hvac_state,
            'fan_state': self.fan_state,
            'heat_setpoint': self.set_point_heat,
            'cool_setpoint': self.set_point_cool,
            'set_point': self.set_point,
            'current_temperature': self.current_temperature,
            'humidity': self.bme280.state()['humidity'],
            'pressure': self.bme280.state()['pressure'],
        }

    def payload(self):
        return json.dumps(self.state())

    def callback(self, **kwargs):
        # system active, should we turn it off?
        logging.info('Checking temperature...temp = {}, heat_setpoint = {}, cool_setpoint = {}, set_point_tolerance = {}'.format(self.temperature, self.set_point_heat, self.set_point_cool, self.set_point_tolerance))
        self.append_tempearture_history()
        
        if self.active and self.mode == HVAC.OFF:
            logging.info('Mode is off, but system is active. Turning HVAC off')
            self.off()

            # reset mode to normal heat
            if self._boosting_heat:
                self.boost_heat(HVAC.OFF)

        elif self.active:
            if self.mode in ['heat', 'aux'] and self.temperature > self.set_point_heat + self.set_point_tolerance:
                # turn hvac off
                logging.info('Temperature is {}. Turning heat off.'.format(self.temperature))
                self.off()

                # reset mode to normal heat
                if self._boosting_heat:
                    self.boost_heat(HVAC.OFF)

            elif self.mode == 'cool' and self.temperature < self.set_point_cool - self.set_point_tolerance:
                # turn hvac off
                logging.info('Temperature is {}. Turning cool off.'.format(self.temperature))
                self.off()

            # should system boost heating with aux heat?
            logging.debug("Checking temperature rate of change...current rate = {}, min rate = {}".format(self.temperature_rate_of_change, self.minimum_temp_rate_of_change))
            if self.mode == HVAC.HEAT and self.temperature_rate_of_change and self.temperature_rate_of_change <= self.minimum_temp_rate_of_change:
                self.boost_heat(HVAC.ON)

        else:
            if self.mode == 'heat' and self.temperature < self.set_point_heat - self.set_point_tolerance:
                # turn hvac on
                logging.info('Temperature is {}. Turning heat on.'.format(self.temperature))
                self.on()
            elif self.mode == 'cool' and self.temperature > self.set_point_cool + self.set_point_tolerance:
                # turn hvac on
                logging.info('Temperature is {}. Turning cool on.'.format(self.temperature))
                self.on()
            # system is inactive, should we turn it on?
        # logging.info('HVAC is {}. Mode is {}. Temperature is {}.'.format(self.active, self.mode, self.temperature))
        MQTT.publish(self.topic, self.payload())

    # def mode_is_changeable(self):
    #     """Can thermostat active mode be chagned?"""
    #     # stop changes from cooling to heat or vice versa while system is running

    #     minutes_since_last_mode_change = (pendulum.now - self.last_mode_change_time).in_minutes()
    #     return not self.active and self.active_time >= self.min_run_time and self.minutes_since_last_mode_chage >= self.min_trigger_cooldown_time

    def set_mode(self, mode):
        logging.info('Changing mode to {}.'.format(mode))
        if mode in HVAC.HEAT_PUMP_MODES:
            self.mode = mode
            self.last_mode_change_time = pendulum.now()
            Config.save()
        else:
            raise HvacException('{} mode is not a valid HVAC mode'.format(mode))

    def _can_change_hvac_state(self):
        """Don't change HVAC state from heat to cool or vice versa if the system is running."""
        if (self.hvac_state == 'cool' and self.mode == 'heat') or (self.hvac_state == 'heat' and self.mode == 'cool'): 
            logging.warn("Don't change between heating and cooling. Doing so may damage your system.")
        elif self.active and self.active_time <= self.min_run_time:
            logging.warn("System needs to run for atleast {} minutes. Only running for {} minutes.".format(self.min_run_time, self.active_time))
        elif not self.active and self.minutes_since_last_hvac_state_change <= self.min_run_time:
            logging.warn("System needs to idle for atleast {} minutes. Only idle for {} minutes.".format(self.min_run_time, self.minutes_since_last_hvac_state_change))
        elif self.minutes_since_last_mode_change <= self.min_trigger_cooldown_time:
            logging.warn("Can only change mode every {} minutes. It's been {} minutes since last change.".format(self.min_trigger_cooldown_time, self.minutes_since_last_mode_change))
        # elif self.mode == self.hvac_state:
        #     logging.info('Ignoring mode change since HVAC is alread in {} mode'.format(self.mode))
        else:
            return True

    def on(self):
        """Helper to turn HVAC and capture useful logging."""
        if self._can_change_hvac_state():
            self.set_state(self.mode, HVAC.ON)
            # TODO verify state was changed and publish result to MQTT
        else:
            logging.warn('Did not activate {}'.format(self.mode))

    def off(self):
        if self._can_change_hvac_state():
            self.set_state(self.mode, HVAC.OFF)
            self.temperature_history = []
        else:
            logging.warn("Did not deactivate {}.".format(self.mode))

    def fan_on(self):
        self.set_state(HVAC.FAN, HVAC.ON)
        logging.info('Turned fan on.')

    def fan_off(self):
        self.set_state(HVAC.FAN, HVAC.OFF)
        logging.info('Turned fan off.')

    def heat_boost_on(self):
        self.set_state(HVAC.BOOST, HVAC.ON)
        self._boosting_heat = HVAC.ON
        logging.info('Heat boost activated.')

    def heat_boost_off(self):
        self.set_state(HVAC.BOOST, HVAC.OFF)
        self._boosting_heat = HVAC.OFF
        logging.info('Heat boost deactivated.')

    def set_fan_mode(self, fan_mode):
        logging.info('Setting fan mode to {}.'.format(fan_mode))
        if fan_mode == HVAC.FAN_ON:
            self.fan_on()
        else:
            self.fan_off()

    def boost_heat(self, boost):
        # manually switching to AUX heat since we don't want to trigger state or mode safety checks.
        # self.mode = HVAC.AUX
        if self._boosting_enabled_switch.state == HVAC.ON:
            if boost == HVAC.ON:
                self.heat_boost_on()
                self._boosting_start_time = pendulum.now()
                # self._boosting_heat = HVAC.ON
            elif boost == HVAC.OFF:
                self.heat_boost_off()
                # reset boost metadata
                self._boosting_start_time = None
                # self.temperature_history = []
                # self._boosting_heat = HVAC.OFF
            else:
                raise HvacException('{} is not a valid boost value. allowed values are [[{},{}]'.format(boost, HVAC.ON, HVAC.OFF))
        # self._boosting_heat = boost
        else:
            logging.info("Boosting disabled by switch.")
    
    def mqtt_ping(self, topic, callback):
        logging.debug("Checing subcription status on topic {}".format(topic))
        response = MQTT.publish(topic, "ping")
        if response != 'pong':
            logging.warn("Not subscribed to topic {}. Resubscribing...".format(topic))
            MQTT.subscribe(topic, callback)

    """
    MQTT subscription callbacks
    """
    @MQTT.pongable
    def mqtt_set_temperature_set_point_callback(self, client, userdata, message):
        try:
            payload = message.payload.decode()
            logging.info("Received temperature set point update request: {}".format(message.payload))
            payload = float(payload)
            if self.mode == HVAC.HEAT:
                self.set_point_heat = payload
            else:
                self.set_point_cool = payload
            Config.save()
        except Exception as e:
            logging.error('Unable to proces message.', e)

        MQTT.publish(self.topic, self.payload())

    @MQTT.pongable
    def mqtt_set_fan_state_callback(self, client, userdata, message):
        try:
            payload = message.payload.decode().lower()
            logging.info("Received fand mode update request: {}".format(payload))
            self.set_fan_mode(payload)
        except Exception as e:
            logging.error('Unable to proces message.', e)

        MQTT.publish(self.topic, self.payload())

    @MQTT.pongable
    def mqtt_set_mode_callback(self, client, userdata, message):
        try:
            payload = message.payload.decode().lower()
            logging.info("Received HVAC mode update request: {}".format(payload))
            self.set_mode(payload)
        except Exception as e:
            logging.error('Unable to proces message.', e)

        MQTT.publish(self.topic, self.payload())

    @MQTT.pongable
    def mqtt_set_aux_mode_callback(self, client, userdata, message):
        try:
            payload = message.payload.decode().lower()
            logging.info("Received aux mode update request: {}".format(payload))
            self.boost_heat(payload)
        except Exception as e:
            logging.error('Unable to proces message.', e)

        MQTT.publish(self.topic, self.payload())


class HVACAuxSwitch(Switch):

    def __init__(self, name, pin, topic, power_state=HVAC.ON, device_class='switch', device_type='generic_switch'):
        super().__init__(name, pin, topic, device_class, device_type)
        self.power_state = power_state

    @property
    def homeassistant_mqtt_config(self):
        config = super(HVACAuxSwitch, self).homeassistant_mqtt_config
        config['value_template'] = "{{ value_json.aux_enabled }}"
        return config

    def setup(self, lazy_setup=True):
        # return super().setup(lazy_setup)
        # don't need any setup
        MQTT.subscribe(self.homeassistant_mqtt_config['command_topic'], self.mqtt_callback)
        # self.on()

    def setup_output(self):
        # don't need to setup GPIO
        pass

    def on(self):
        self.power_state = HVAC.ON

    def off(self):
        self.power_state = HVAC.OFF

    def state(self):
        return self.power_state
    


# coding=utf-8
import Adafruit_DHT as dht
import json
from rpi2mqtt.mqtt import MQTT as mqtt
from rpi2mqtt.base import Sensor, SensorGroup, sensor
import logging
import os
import glob
import smbus2
import bme280
import re


class DHT(object):

    def __init__(self, pin, topic, name, device_class, dht_type):
        self.type = dht_type
        self.pin = pin
        self.topic = topic
        self.name = name
        self.device_class = device_class
        self.setup()

    def read(self, scale='F'):
        self.humidity, self.temperature = dht.read_retry(22, self.pin)
        #print humidity
        #print temperature

        if scale == 'F':
            return json.dumps({'humidity': self._humidity, 'temperature': self.temperature_F})
        else:
            return json.dumps({'humidity': self._humidity, 'temperature': self.temperature})

    @property
    def temperature_F(self):
        try:
            return round(self.temperature_C * 1.8 + 32.0, 1)
        except:
            pass

    @property
    def temperature_C(self):
        try:
            return round(self.temperature, 1)
        except:
            pass

    @property
    def _humidity(self):
        try:
            return round(self.humidity, 3)
        except:
            pass

    def setup(self):
        # config = json.dumps({'name': self.name, 'device_class': self.device_class})

        device_config = {'name': "Laundry Room Climate",
                         'identifiers': self.name,
                         'sw_version': 'rpi2mqtt',
                         'model': "DHT 22",
                         'manufacturer': 'Generic'}

        config = json.dumps({'name': self.name + '_temperature',
                             'device_class': 'temperature',
                             'unit_of_measurement': '°F',
                             'value_template': "{{ value_json.temperature }}",
                             'unique_id': self.name + '_temperature_rpi2mqtt',
                             'state_topic': self.topic,
                             "json_attributes_topic": self.topic,
                             'device': device_config})

        mqtt.publish('homeassistant/sensor/{}_{}/config'.format(self.name, 'temp'), config)

        config = json.dumps({'name': self.name + '_humidity',
                             'device_class': 'humidity',
                             'json_attributes_topic': self.topic,
                             'unit_of_measurement': '%',
                             'value_template': "{{ value_json.humidity }}",
                             'unique_id': self.name + '_humidity_rpi2mqtt',
                             'state_topic': self.topic,
                             'device': device_config})

        mqtt.publish('homeassistant/sensor/{}_{}/config'.format(self.name, 'humidity'), config)

    def state(self):
        return self.read()

    def payload(self):
        # return json.dumps({'state': self.state()})
        # return json.dumps(self.state())
        return self.state()

    def callback(self, **kwargs):
        mqtt.publish(self.topic, self.payload())


class GenericTemperature(Sensor):
    @property
    def homeassistant_mqtt_config(self):
        config = super(GenericTemperature, self).homeassistant_mqtt_config
        config['value_template'] = "{{ value_json.temperature }}"
        config['unit_of_measurement'] = '°F'
        return config

    def setup(self):
        pass

    def state(self):
        pass

class GenericHumidity(Sensor):
    @property
    def homeassistant_mqtt_config(self):
        config = super(GenericHumidity, self).homeassistant_mqtt_config
        config['value_template'] = "{{ value_json.humidity }}"
        config['unit_of_measurement'] = '%'
        return config

    def setup(self):
        pass

    def state(self):
        pass

class GenericPressure(Sensor):
    @property
    def homeassistant_mqtt_config(self):
        config = super(GenericPressure, self).homeassistant_mqtt_config
        config['value_template'] = "{{ value_json.pressure }}"
        return config

    def setup(self):
        pass

    def state(self):
        pass

class BME280(SensorGroup):

    def __init__(self, name, topic, **kwargs):
        super(BME280, self).__init__(name, None, topic, 'temperature/humidity/pressure', 'BME280', **kwargs)
        self.port = 1
        self.address = 0x76
        self.bus = smbus2.SMBus(self.port)
        self.calibration_params = bme280.load_calibration_params(self.bus, self.address)
        self.topic = topic
        self.device_type = 'BME280'
        self.setup_temperature()
        self.setup_humidity()
        self.setup_pressure()


        # the sample method will take a single reading and return a
        # compensated_reading object
        # data = bme280.sample(bus, address, calibration_params)

        # TODO sensors with multiple reading types can be supported with publishing one message
        # but each reading type must be setup seperately with different names and value_json attributes.
    
    # @sensor
    def setup_temperature(self):
        sensor = GenericTemperature(self.name, None, self.topic, 'temperature', self.device_type)
        self.sensors.append(sensor)
    
    # @sensor
    def setup_humidity(self):
        sensor = GenericHumidity(self.name, None, self.topic, 'humidity', self.device_type)
        self.sensors.append(sensor)

    # @sensor
    def setup_pressure(self):
        sensor = GenericPressure(self.name, None, self.topic, 'pressure', self.device_type)
        self.sensors.append(sensor)

    def state(self):
        data = bme280.sample(self.bus, self.address, self.calibration_params)
        return {'id': str(data.id),
            'timestamp': str(data.timestamp),
            'temperature': data.temperature * 1.8 + 32,
            'pressure': data.pressure,
            'humidity': data.humidity,
            }

    def payload(self):
        return json.dumps(self.state())


class OneWire(Sensor):
    """Must enable one wire interface on Raspberry Pi and load modprobe w1-gpio and w1-therm drivers."""
    BASE_DIR = '/sys/bus/w1/devices/'
    # These tow lines mount the device:
    try:
        os.system('modprobe w1-gpio')
        os.system('modprobe w1-therm')
    except:
        logging.warn('Unable to load w1-gpio and w1-therm.')

    TEMP_REGEX = re.compile('.*? t=(\d*)')

    def __init__(self, name, topic, **kwargs):
        self.devices = {}
        self.temperature = None
        self.setup()
        super(OneWire, self).__init__(name, None, topic, 'temperature', 'One wire', **kwargs)

    def setup(self):
        for device in glob.glob( OneWire.BASE_DIR + '**/w1_slave'):
            w1_id = device.split('/')[-2]
            self.devices[w1_id] = device
        return True

    def state(self):
        for device, filename in self.devices.items():
            with open(filename, 'r') as f:
                self.temperature = OneWire.parse_one_wire_file(device, f.read())

            return self.temperature_F

    def payload(self):
        return json.dumps({
            'temperature': self.state(),
            'state': self.state()
        })
    
    @property
    def temperature_F(self):
        try:
            return round(self.temperature * 1.8 + 32.0, 1)
        except:
            pass

    @property
    def temperature_C(self):
        try:
            return round(self.temperature, 1)
        except:
            pass

    @staticmethod
    def parse_one_wire_file(device, text):
        match = OneWire.TEMP_REGEX.search(text)
        if text:
            return float(match.groups()[0]) / 1000.0
        else:
            logging.warn('Unable to parse temperature for one wire device {}.'.format(device))



import RPi.GPIO as GPIO
from rpi2mqtt.base import Sensor
from rpi2mqtt.mqtt import MQTT as mqtt
import json
import logging


class ReedSwitch(Sensor):
    """
    Extends simple binary sensor by adding configuration for normally open or normally closed reed switches.
    """

    def __init__(self, name, pin, topic, normally_open, device_class=None):
        super(ReedSwitch, self).__init__(name, pin, topic, device_class, 'reed_switch')
        self.normally_open = normally_open
        self.setup()

    def setup(self):
        """
        Setup GPIO pin to read input value.
        :return: None
        """
        # device_config = {'name': self.name,
        #                  'identifiers': self.name,
        #                  'sw_version': 'rpi2mqtt',
        #                  'model': self.device_model,
        #                  'manufacturer': 'Generic'}

        # config = json.dumps({'name': self.name + '_' + self.device_model,
        #                      'device_class': self.device_class,
        #                      'value_template': "{{ value_json.state }}",
        #                      'unique_id': self.name + '_' + self.device_model + '_rpi2mqtt',
        #                      'state_topic': self.topic,
        #                      "json_attributes_topic": self.topic + '/state',
        #                     #  "command_topic": self.topic + '/set',
        #                      'device': device_config})

        # mqtt.publish('homeassistant/{}/{}_{}/config'.format(self.device_class, self.name, self.device_model), config)
        # logging.debug("Published MQTT discovery config to homeassistant/{}/{}_{}/config.format(self.device_class, self.name, self.device_model)")
        GPIO.setmode(GPIO.BCM)
        # g.setup(self.pin, g.OUT)
        # mqtt.subscribe(self.topic + '/set', self.mqtt_callback)

        if self.normally_open:
            mode = GPIO.PUD_UP
        else:
            mode = GPIO.PUD_DOWN

        GPIO.setup(self.pin, GPIO.IN, pull_up_down=mode)
        logging.debug('Reed Switch {} configured as input on GPIO{} witn pull_up_down set to {}'.format(self.name, self.pin, mode))

    def state(self):
        state = GPIO.input(self.pin) 
        logging.debug("Reed Switch {}: GPIO{} state is {}".format(self.name, self.pin, state))
        if state == 1:
            return "ON"
        else:
            return "OFF"

    def payload(self):
        return json.dumps({'state': self.state()})

    def callback(self, *args):
        mqtt.publish(self.topic, self.payload())


class BinarySensor(Sensor):

    def __init__(self, name, pin, topic ):
        super(BinarySensor, self).__init__(name, pin, topic, None, 'binary')
        self.setup()

    def setup(self):
        pass

    def state(self):
        pass



from rpi2mqtt.mqtt import MQTT as mqtt
import json
import logging
from rpi2mqtt.version import __version__
import RPi.GPIO as GPIO

# logging.basicConfig(level=logging.INFO)


SENSORS = {}
def sensor(func):
    def wrap(*args):
        if args.name not in SENSORS:
            SENSORS[args.name] = [func]
        else:
            SENSORS[args.name].append(func)
    return wrap

GPIO.setmode(GPIO.BCM)

class Sensor(object):

    BINARY_SENSORS = ['reed', 'binary_sensor']

    def __init__(self, name, pin, topic, device_class, device_model, **kwargs):
        self.name = name
        self.pin = pin
        self.topic = topic
        self.device_class = device_class
        self.device_model = device_model
        self.publish_mqtt_discovery()
        

    @property
    def device_config(self):
        return {'name': self.name,
                'identifiers': self.name,
                'sw_version': 'rpi2mqtt {}'.format(__version__),
                'model': self.device_model,
                'manufacturer': 'Generic'}

    @property
    def homeassistant_mqtt_config(self):
        return {'name': '{}_{}'.format(self.name, self.device_class),
                'device_class': self.device_class,
                'value_template': "{{ value_json.state }}",
                'unique_id': '{}_{}_{}_rpi2mqtt'.format(self.name, self.device_model, self.device_class),
                'state_topic': self.topic,
                "json_attributes_topic": self.topic,
                'device': self.device_config}

    @property
    def homeassistant_mqtt_config_json(self):
        return json.dumps(self.homeassistant_mqtt_config)

    @property
    def homeassistant_mqtt_config_topic(self):
        homeassistant_sensor_type = 'sensor'
        if self.device_class in Sensor.BINARY_SENSORS:
            homeassistant_sensor_type = 'binary_sensor'
        elif self.device_class in ['switch', 'off', 'heat', 'cool', 'aux']:
            homeassistant_sensor_type = 'switch'
        elif self.device_class == 'climate':
            homeassistant_sensor_type = 'climate'

        return 'homeassistant/{}/{}_{}/config'.format(homeassistant_sensor_type, self.name, self.device_class)

    def publish_mqtt_discovery(self):

        mqtt.publish(self.homeassistant_mqtt_config_topic, self.homeassistant_mqtt_config_json)
        logging.debug("Published MQTT discovery config to {}".format(self.homeassistant_mqtt_config_topic))

    def setup(self):
        raise NotImplementedError("Setup method is required.")

    def state(self):
        raise NotImplementedError("State method is required.")

    def payload(self):
        return json.dumps({'state': self.state()})

    def callback(self, **kwargs):
        mqtt.publish(self.topic, self.payload())


class SensorGroup(Sensor):
    """Group of multiple sensors e.g. DHT22 or BME280.

    Sensors with multiple capabilities (i.e. temperature, humididty, pressure, presenece, etc.) must be setup as
    individual sesnsors in home assistant. Updating the state however, is done with a single message to the sensor's
    state topic.

    Attributes:
        name (str): Sensor name
        pin (int): Sensor GPIO pin (BCM pin nanme)
        topic (str): Base topic name. This is prepended to '/state' and '/config' to create respective topics in HA.
        device_class (str): Home Assistant device class for this sensor.
        device_type (str): Type of sensor. e.g. DHT22, Reed Switch, etc. 
    """
    def __init__(self, name, pin, topic, device_class, device_type):
        self.name = name
        self.pin = pin
        self.topic = topic
        self.device_class = device_class
        self.device_type = device_type
        self.sensors = []

    def setup(self):
        for sensor in self.sensors:
            sensor.setup()

    # def state(self):
    #     raise NotImplementedError('State method is required.')

    # def payload(self): 
    #     raise NotImplementedError('Payload method is required.')

    # def callback(self):
    #     raise NotImplementedError('Callback method is required.')





import logging
import sys
import yaml
# from dotmap import DotMap
import pathlib
from dataclasses import dataclass, asdict
# from typing import Optional


logging.basicConfig(stream=sys.stdout, format='%(asctime)s:%(levelname)s:%(message)s')


@dataclass
class SensorConfig:
    type: str
    name: str
    topic: str
    pin: int = None
    # device_class: str = None
    # binary sensor options
    normally_open: bool = None
    # hestia pi thermostat options
    mode: str = None
    heat_setpoint: float = None
    cool_setpoint: float = None
    aux_enabled: str = None
    set_point_tolerance: float = None
    min_run_time: int = None
    min_trigger_cooldown_time: int = None
    temperature_history_period: int = None
    minimum_temp_rate_of_change: int = None
    dry_run: bool = None
    # ibeacon options
    beacon_uuid: str = None
    beacon_away_timeout: int = None


@dataclass
class MqttConfig:
    host: str
    port: int
    username: str = None
    password: str = None
    ca_cert: str = None
    retries: int = 3


@dataclass
class Conf:
    # filename: str
    mqtt: MqttConfig
    polling_interval: int
    log_level: str
    sensors: list


class Config():
    """Config holder singleton. Use Config.get_instance() in submodules to access. Initial call must include 'filename'.

    Attributes:
        [type]: [description]
    """
    _config = None
    _filename = None

    @classmethod
    def get_instance(cls, filename=None):
        if filename:
            cls._filename = filename

        if not cls._config and cls._filename:
            with open(cls._filename, 'r') as f:
                # cls._config = DotMap(yaml.safe_load(f))
                Config.load(yaml.safe_load(f))
                config_filename = pathlib.Path(f.name).absolute()
                logging.info("Loaded config file {}".format(config_filename))
            Config.set_log_level(cls._config.log_level)
        return cls._config

    @classmethod
    def load(cls, config):
        cls._config = Conf(**config)
        cls._config.mqtt = MqttConfig(**cls._config.mqtt)
        cls._config.sensors = [SensorConfig(**sensor) for sensor in cls._config.sensors]

    @classmethod
    def to_dict(cls):
        _cfg = asdict(cls._config)
        sensors = []
        for sensor in _cfg['sensors']:
            sensors.append({k:v for k,v in sensor.items() if v})
        _cfg['sensors'] = sensors
        return _cfg

    @staticmethod
    def set_log_level(level):
        _level = logging.WARN
        if level == 'info':
            _level = logging.INFO
        elif level == 'debug':
            _level = logging.DEBUG
        elif level == 'error':
            _level = logging.ERROR
        logger = logging.getLogger().setLevel(_level)

    @classmethod
    def save(cls):
        logging.info("Saving config file...")
        with open(cls._filename, 'w') as f:
            yaml.dump(cls._config.to_dict(), f)


def generate_config(config_filename):
    with open(config_filename, 'w') as f: 
        default_config = {
            "log_level": 'warn',
            'polling_interval': 300,
            'mqtt': {
                'host': 'hostname',
                'port': 1883,
                'username': 'mqtt_user',
                'password': 'mqtt_password',
                'retries': 3,
                'ca_cert': '/path/to/certificate'
            },
            'sensors': [
                {'name': 'sensor_name', 'type': 'sensor_type', 'pin': 'gpio pin number'}
            ]}
        f.write(yaml.dump(default_config))




import paho.mqtt.publish as _mqtt
# import paho.mqtt.subscribe as mqtt_sub
from paho.mqtt.client import Client
from rpi2mqtt.config import Config
# import traceback
import logging
# import sys
import pendulum


# logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)


def on_message(mqttc, obj, msg):
    logging.info("Recieved: " + msg.topic + " " + str(msg.qos) + " " + str(msg.payload))


def on_subscribe(mqttc, obj, mid, granted_qos):
    logging.info("Subscribed to " + str(mid) + " " + str(granted_qos))


class Subscription():
    def __init__(self, topic, callback):
        self.topic = topic
        self.callback = callback
        self.last_ping = pendulum.now()


class MQTT():
    
    client = None
    subscribed_topics = None
    config = None

    @classmethod
    def publish(cls, topic, payload, cnt=1):
        try:
            logging.info("Pushlishing to topic {}: | attempt: {} | message: {}".format(topic, cnt, payload))
            if cnt <= cls.config.mqtt.retries:
                if payload == 'pong':
                    cls.subscribed_topics[topic].last_ping = pendulum.now()

                _mqtt.single(topic, payload, 
                            hostname=cls.config.mqtt.host, 
                            port=cls.config.mqtt.port,
                            auth={'username': cls.config.mqtt.username, 'password': cls.config.mqtt.password},
                            tls={'ca_certs': cls.config.mqtt.ca_cert},
                            retain=True)
        except Exception as e:
            logging.exception("Error publishing message.", e)
            cnt += 1
            cls.publish(topic, payload, cnt)

    @classmethod
    def setup(cls):
        cls.client = Client()
        cls.subscribed_topics = {}
        cls.config = Config.get_instance()
        cls.client.tls_set(ca_certs=cls.config.mqtt.ca_cert) #, certfile=None, keyfile=None, cert_reqs=cert_required, tls_version=tlsVersion)

        # if args.insecure:
        #     self.client.tls_insecure_set(True)

        if cls.config.mqtt.username or cls.config.mqtt.password:
            cls.client.username_pw_set(cls.config.mqtt.username, cls.config.mqtt.password)

        logging.info("Connecting to " + cls.config.mqtt.host + " port:" + str(cls.config.mqtt.port))
        cls.client.connect(cls.config.mqtt.host, cls.config.mqtt.port, 60)
        logging.info("Successfully connected to {} port:{}".format(cls.config.mqtt.host, str(cls.config.mqtt.port)))

        cls.client.loop_start()

        cls.client.on_subscribe = on_subscribe
        cls.client.on_message = on_message



    @classmethod
    def subscribe(cls, topic, callback):
        logging.info("Subscribing to topic %s", topic)
        res = cls.client.subscribe(topic)
        logging.info('Subscription result = {}'.format(res))
        cls.client.message_callback_add(topic, callback)
        cls.subscribed_topics[topic] = Subscription(topic, callback)
        return res

    @classmethod
    def ping_subscriptions(cls):
        for topic, sub in cls.subscribed_topics.items():
            logging.debug("Checing subcription status on topic {}".format(topic))
            response = MQTT.publish(topic, "ping")
            last_seen = pendulum.now() - cls.subscribed_topics[topic].last_ping
            if last_seen and last_seen.seconds > cls.config.polling_interval * 2:
                logging.warn("Not subscribed to topic {}. Resubscribing...".format(topic))
                MQTT.subscribe(topic, sub.callback)

    @classmethod
    def refresh_subscriptions(cls):
        for topic, sub in cls.subscribed_topics.items():
            MQTT.client.unsubscribe(topic)
            MQTT.subscribe(sub.topic, sub.callback)

    @staticmethod
    def pongable(func):
        # def decorator_wrapper(func):
        def wrapper(self, client, userdata, message):
            logging.debug('Received message {} on topic {}'.format(message.payload.decode(), message.topic))
            payload = message.payload.decode()
            if payload == 'ping':
                MQTT.publish(message.topic, "pong") 
                return
            elif payload == 'pong':
                return
            return func(self, client, userdata, message)
        return wrapper
        # return decorator_wrapper


# import asyncio
import logging
import traceback
import argparse
import subprocess
import sys
import schedule

from rpi2mqtt.config import Config
from rpi2mqtt.binary import *
from rpi2mqtt.temperature import *
from rpi2mqtt.ibeacon import Scanner
from rpi2mqtt.switch import Switch
from rpi2mqtt.thermostat import HestiaPi
import time

try:
    from beacontools import BeaconScanner, IBeaconFilter
except:
    print("Unable to load beacontools")


# setup CLI parser
parser = argparse.ArgumentParser()
parser.add_argument("-c", "--config",
                help="Path to config.yaml")

parser.add_argument('-d', '--dry-run', 
                help='Test drive config without triggering callbacks.')

parser.add_argument('--generate-config',
                help="Generate config.yaml template.",
                action='store_true')

parser.add_argument('--install-service',
                help='Install rpi2mqtt as systemd service.',
                action='store_true')

parser.add_argument('--install-user-service',
                help='Install rpi2mqtt as user systemd service.',
                action='store_true')


def main():
    config = None
    args = parser.parse_args() 

    if args.generate_config:
        Config.generate_config('config.yaml')
        sys.exit(0)

    if args.install_service:
        username = input("User to run service as [pi]: ") or 'pi'
        config_path = input("Enter full path to config.yaml: ")
        # _path = input("Path rpi2mqtt executable (run `which rpi2mqtt`): ")
        _path = subprocess.check_output(['which', 'rpi2mqtt']).decode().strip()
        install_service(username, _path, config_path, 'system')
        sys.exit(0)

    if args.install_user_service:
        username = input("User to run service as [pi]: ") or 'pi'
        config_path = input("Enter full path to config.yaml: ")
        # _path = input("Path rpi2mqtt executable (run `which rpi2mqtt`): ")
        _path = subprocess.check_output(['which', 'rpi2mqtt']).decode().strip()
        install_service(username, _path, config_path, 'user')
        sys.exit(0)

    scanner = None

    if args.config:
        config = Config.get_instance(filename=args.config)

    if not config:
        logging.error("No configuration file present.")
        sys.exit(1)

    # start MQTT client
    from rpi2mqtt.mqtt import MQTT
    MQTT.setup()

    sensor_list = []
    if len(config.sensors) > 0:
        for sensor in config.sensors:
            s = None
            if sensor.type == 'dht22':
                s = DHT(sensor.pin, sensor.topic, sensor.name, 'sensor', sensor.type)
            elif sensor.type == 'ibeacon':
                s = Scanner(sensor.name, sensor.topic, sensor.uuid, sensor.away_timeout)
            elif sensor.type == 'switch':
                s = Switch(sensor.name, sensor.pin, sensor.topic)
            elif sensor.type == 'reed':
                s = ReedSwitch(sensor.name, sensor.pin, sensor.topic, sensor.normally_open, sensor.get('device_type'))
            elif sensor.type == 'bme280':
                s = BME280(sensor.name, sensor.topic)
            elif sensor.type == 'hestiapi':
                s = HestiaPi(**sensor)
            elif sensor.type == 'onewire':
                s = OneWire(sensor.name, sensor.topic)
            else:
                logging.warn('Sensor {} found in config, but was not setup.'.format(sensor.name))
            if s:
                sensor_list.append(s)

        try:
            scanner = BeaconScanner(sensor_list[1].process_ble_update) # TODO update to search sensor list and setup scanner accordingly.
            scanner.start()
        except:
            logging.error("Beacon scanner did not start")
    else:
        logging.warn("No sensors defined in {}".format(args.config))

    schedule.every().day.at("01:00").do(MQTT.refresh_subscriptions)

    try:
        while True:

            for sensor in sensor_list:
                sensor.callback()

            time.sleep(config.polling_interval)
            MQTT.ping_subscriptions()
            schedule.run_pending()

    except:
        traceback.print_exc()
        MQTT.client.loop_stop()

        if scanner:
            scanner.stop()


def install_service(username, _path, config_path, _type):
    template = """[Unit]
Description=rpi2mqtt Service
After=network-online.target

[Service]
# replace user with an existing system user
Restart=on-failure
User={username}
ExecStart={_path} -c {config_path}

[Install]
WantedBy=multi-user.target
    """.format(username=username, _path=_path, config_path=config_path)
    # return template
    if _type == 'user':
        filename = '~/.config/systemd/user/rpi2mqtt.service'
    else:
        filename = '/etc/systemd/system/rpi2mqtt.service'

    with open(filename, 'w') as f:
        f.write(template)

if __name__ == '__main__':
    main()


from rpi2mqtt.base import Sensor
from rpi2mqtt.mqtt import MQTT as mqtt
import json
from datetime import datetime, timedelta
import RPi.GPIO as g
import logging


class BasicSwitch(Sensor):
    """Basic switch setup to control GPIO pins and read their state

    Args:
        Sensor ([type]): [description]
    """
    def __init__(self, name, pin, topic, device_class='switch', device_type='generic_switch'):
        super(BasicSwitch, self).__init__(name, pin, topic, device_class, device_type)
        self.power_state = 'OFF'
        self.last_seen = datetime.now()
        # self.setup()
        # g.setmode(g.BCM)

    def setup(self, lazy_setup=True):
        """
        Setup Home Assistant MQTT discover for ibeacons.
        :return: None
        """
        # setup GPIO
        g.setmode(g.BCM)
        if not type(self.pin) == list:
            self.pin = [self.pin]

        for pin in self.pin:
            g.setup(pin, g.IN)
        
        if not lazy_setup:
            self.setup_output()

    def setup_output(self):
        logging.info("Setting pins {} to ouptut.".format(self.pin))
        g.setup(self.pin, g.OUT, initial=g.LOW)

    def on(self):
        try:
            g.output(self.pin, g.HIGH)
        except RuntimeError as e:
            logging.info("Switch output not configured yet. Setting up pins {}".format(self.pin))
            self.setup_output()
            g.output(self.pin, g.HIGH)
        self.power_state = 'ON'

    def off(self):
        try:
            g.output(self.pin, g.LOW)
        except RuntimeError as e:
            logging.info("Switch output not configured yet. Setting up pins {}".format(self.pin))
            self.setup_output()
            g.output(self.pin, g.LOW)
        self.power_state = 'OFF'

    def toggle(self):
        if self.power_state == 'ON':  # and self.last_seen + timedelta(seconds=self.away_timeout) < datetime.now():
            self.off()
        else:
            self.on()

    def state(self):
        # read output pin state
        pin_state = g.input(self.pin)

        # convert to home assistant on/off state defaults
        # https://www.home-assistant.io/integrations/switch.mqtt/#state_on
        if pin_state == 1:
            self.power_state = 'ON'
        elif pin_state == 0:
            self.power_state = 'OFF'
        else:
            self.power_state = 'INVALID'

        return self.power_state

    def payload(self):
        return json.dumps({'power_state': self.state()})

    def publish_mqtt_discovery(self):
        pass




class Switch(Sensor):

    def __init__(self, name, pin, topic, device_class='switch', device_type='generic_switch'):
        super(Switch, self).__init__(name, pin, topic, device_class, device_type)
        self.power_state = 'OFF'
        self.last_seen = datetime.now()
        self.setup()


    @property
    def homeassistant_mqtt_config(self):
        config = super(Switch, self).homeassistant_mqtt_config
        config['value_template'] = "{{ value_json.power_state }}"
        config['command_topic'] = self.topic + '/set'
        del config['device_class']
        return config

    def setup(self, lazy_setup=True):
        """
        Setup Home Assistant MQTT discover for ibeacons.
        :return: None
        """
        # setup GPIO
        g.setmode(g.BCM)
        if not type(self.pin) == list:
            self.pin = [self.pin]

        for pin in self.pin:
            g.setup(pin, g.IN)

        # for pin in self.pin:
        if not lazy_setup:
            self.setup_output()
        mqtt.subscribe(self.homeassistant_mqtt_config['command_topic'], self.mqtt_callback)

    def setup_output(self):
        logging.debug("Setting pins {} to ouptut.".format(self.pin))
        g.setup(self.pin, g.OUT, initial=g.LOW)

    def on(self):
        try:
            g.output(self.pin, g.HIGH)
        except RuntimeError as e:
            logging.info("Switch output not configured yet. Setting up pins {}".format(self.pin))
            self.setup_output()
            g.output(self.pin, g.HIGH)
        self.power_state = 'ON'

    def off(self):
        try:
            g.output(self.pin, g.LOW)
        except RuntimeError as e:
            logging.info("Switch output not configured yet. Setting up pins {}".format(self.pin))
            self.setup_output()
            g.output(self.pin, g.LOW)
        self.power_state = 'OFF'

    def toggle(self):
        if self.power_state == 'ON':  # and self.last_seen + timedelta(seconds=self.away_timeout) < datetime.now():
            self.off()
        else:
            self.on()

    def state(self):
        # read output pin state
        # TODO refactor switch into single pin & multi pin classes
        pin_state = 0
        for _pin in self.pin:
            pin_state += g.input(_pin)

        # convert to home assistant on/off state defaults
        # https://www.home-assistant.io/integrations/switch.mqtt/#state_on
        if pin_state == 1:
            self.power_state = 'ON'
        elif pin_state == 0:
            self.power_state = 'OFF'
        else:
            self.power_state = 'INVALID'

        return self.power_state

    def payload(self):
        return json.dumps({'power_state': self.state()})

    # def callback(self, *args):
    #     mqtt.publish(self.topic, self.payload())
    @mqtt.pongable
    def mqtt_callback(self, client, userdata, message):
        try:
            # print(message)
            # logging.info("Received command message: {}".format(payload))
            payload = message.payload.decode()
            logging.info("Received command message: {} on topic ".format(payload, message.topic))
            if payload == 'ON':
                self.on()
            elif payload == 'OFF':
                self.off()
        except Exception as e:
            logging.error('Unable to proces message.', e)

        mqtt.publish(self.topic, self.payload())

    